/* ***** BEGIN LICENSE BLOCK *****
 * 
 * Copyright (c) 1997-2008 Aliasource - 2009-2011 Groupe LINAGORA
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of the
 *  License, (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 * 
 *  http://www.obm.org/                                              
 * 
 * ***** END LICENSE BLOCK ***** */

const Cc = Components.classes;
const Ci = Components.interfaces;


/**
 * main object to access to OBM calendars
 * @tag NEW_OBM_CALENDAR_MANAGER
 */
function ObmCalendarsManager () {
	// Init logger
	this.logger = Cc["@obm.org/calendar/logger;1"].createInstance(Ci.obmILogger);
	var writer = Cc["@obm.org/calendar/writer;1?type=unixfile"].getService(Ci.obmILoggerWriter);
	this.logger.setLevelWriter(this.logger.LOAD_LEVEL, writer);
	this.logger.setHeader("ObmCalendarsManager.js: ");
	this.logger.debug("new ObmCalendarsManager()", false, false);

	this.prefService = Components.classes["@mozilla.org/preferences-service;1"]
							.getService(Components.interfaces.nsIPrefBranch);
	this.prefService.QueryInterface(Components.interfaces.nsIPrefService);

	this.calendarManager = Cc["@mozilla.org/calendar/manager;1"]
								.getService(Ci.calICalendarManager);

	var scriptLoader = Cc["@mozilla.org/moz/jssubscript-loader;1"].createInstance(Ci.mozIJSSubScriptLoader);
	if (typeof(UtilsOBM) == "undefined") {
		scriptLoader.loadSubScript("chrome://obm-extension/content/calendar/UtilsOBM.js");
	}
};

ObmCalendarsManager.prototype = {

	QueryInterface: function(aIID) {
		if (!aIID.equals(Ci.nsISupports) && !aIID.equals(Ci.obmICalendarsManager)) {
			throw Cr.NS_ERROR_NO_INTERFACE;
		}
		return this;
	},

	_getCharPref : function (aPrefName, aDefaultValue) {
		var prefValue;
		try {
			prefValue = this.prefService.getCharPref(aPrefName);
		} catch (e) {
			prefValue = aDefaultValue | "";
		}
		return prefValue;
	},

	_setCharPref : function (aPrefName, aValue) {
		this.prefService.setCharPref(aPrefName, aValue);
		this.prefService.savePrefFile(null);
	},

	_getNextLocalCalendarID : function () {
		var calendars = this.calendarManager.getCalendars({});

		// FIXME n'importe quoi !!!
		var minID = 20;

		for each ( var calendar in calendars ) {
		  if ( calendar.uri ) {
			var calendarID = calendar.uri.spec.match(/id=(\d+)/);
			if ( calendarID != null ) {
			  if ( Number(calendarID[1]) >= minID ) {
				minID = Number(calendarID[1]) + 1;
			  }
			}
		  }
		}

		return minID;
	},

	getObmCalendar : function (aCalInfoUid) {
		var calendars = this.calendarManager.getCalendars({});
		for each(var calendar in calendars) {
			if (calendar.type == "obm" && calMgr.getCalendarPref_(calendar, "X-OBM-NAME") == aCalInfoUid) {
				return calendar;
			}
		}
		return null;
	},

	subscribeObmCalendar: function (aUid, aName, aEmail, aReadOnly) {

		var name = "[OBM] " + aName;
		var color = UtilsOBM.getRandomColor();
		//dump("_subscribeTo: " + aUid + ", " + name + ", " +  aEmail + ", " + color + ", " + aReadOnly + "\n");

		this.prefService.setBoolPref(
		  "extensions.obm.changes.calICalendar."
		  + aUid
		  + ".readOnly", aReadOnly
		);

		var uri = Cc["@mozilla.org/network/simple-uri;1"].createInstance(Ci.nsIURI);
		uri.spec = "moz-profile-calendar://?id=" + this._getNextLocalCalendarID() + ";" + aUid;

		var newCalendar = this.calendarManager.createCalendar("obm", uri);
		this.calendarManager.registerCalendar(newCalendar);

		newCalendar.name = name;
		newCalendar.setProperty("color", color.toString());

		this.calendarManager.setCalendarPref_(newCalendar, "X-OBM-NAME", aUid.toString()); 
		this.calendarManager.setCalendarPref_(newCalendar, 'X-OBM-EMAIL', aEmail.toString());

		this._setCharPref("extensions.obm.changes."
				+ Ci.calICalendar + "."
				+ aUid + "."
				+ "lastSync",
				"1970-01-01T00:00:00Z");
	},

	updateObmCalendar: function(aCalendar, aUid, aName, aEmail, aReadOnly) {

		var name = "[OBM] " + aName;
		aCalendar.name = name;

		//dump("_updateCalendar: " + aUid + ", " + name + ", " +  aEmail + ", " + aReadOnly + "\n");

		this.prefService.setBoolPref(
		  "extensions.obm.changes.calICalendar."
		  + aUid
		  + ".readOnly", aReadOnly
		);

		this.calendarManager.setCalendarPref_(aCalendar, "X-OBM-NAME", aUid.toString()); 
		this.calendarManager.setCalendarPref_(aCalendar, 'X-OBM-EMAIL', aEmail.toString());
	}
};


var ObmCalendarsModule = {
	mCID: Components.ID("{a55551b2-2bea-406c-9b6b-f9be098ed178}"),
	mContractID: "@obm.org/calendarsmanager;1",

	registerSelf: function(compMgr, fileSpec, location, type) {
		compMgr = compMgr.QueryInterface(Ci.nsIComponentRegistrar);
		compMgr.registerFactoryLocation(this.mCID, "OBM Calendars Manager", this.mContractID, fileSpec, location, type);
	},
	getClassObject: function(compMgr, cid, iid) {
		if (!cid.equals(this.mCID)) {
			throw Cr.NS_ERROR_NO_INTERFACE;
		}
		if (!iid.equals(Ci.nsIFactory)) {
			throw Cr.NS_ERROR_NOT_IMPLEMENTED;
		}
		return this.mFactory;
	},
	mFactory: {
		createInstance: function(outer, iid) {
			if (outer != null) {
				throw Cr.NS_ERROR_NO_AGGREGATION;
			}
			return (new ObmCalendarsManager()).QueryInterface(iid);
		}
	},

	canUnload: function(compMgr) {
		return true;
	}
};

function NSGetModule(compMgr, fileSpec) {
	return ObmCalendarsModule;
}


