/* ***** BEGIN LICENSE BLOCK *****
 * 
 * Copyright (c) 1997-2008 Aliasource - Groupe LINAGORA
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of the
 *  License, (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 * 
 *  http://www.obm.org/                                              
 * 
 * ***** END LICENSE BLOCK ***** */
const Cc = Components.classes;
const Ci = Components.interfaces;
const Cr = Components.results;

function QuitObserver() {
	if ( this.observerAdded ) {
		return;
	}
	var observerService = Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService);
	observerService.addObserver(this, "quit-application", false);
	this.observerAdded = true;  
}

QuitObserver.prototype = {
	//
	// nsISupports interface
	//

	QueryInterface: function(aIID) {
		if (!aIID.equals(Ci.nsISupports) &&
		!aIID.equals(Ci.nsIObserver)) {
			throw Cr.NS_ERROR_NO_INTERFACE;
		}
		return this;
	},

	//
	// nsIObserver interface
	//
	observe: function(aSubject, aTopic, aData) {
		if (aTopic != "quit-application") {
			dump("observe:"+aSubject+","+aTopic+"\n");
			return;
		}

		if (typeof(UtilsOBM) == "undefined") {
			var scriptLoader = Cc["@mozilla.org/moz/jssubscript-loader;1"].createInstance(Ci.mozIJSSubScriptLoader);
			scriptLoader.loadSubScript("chrome://obm-extension/content/calendar/UtilsOBM.js");
		}

		// TODO : reset pref for obm-sync alert support:
		//UtilsOBM.unsetPref("extensions.obm.sync-supports-alerts");

		var username = {};
		var password = {};
		if ( !UtilsOBM.getOBMCredentials(username, password)
		|| username.value == "" || password.value == "" ) {
			return;
		}

		var pref = UtilsOBM.getBoolPref("extensions.obm.syncOnQuit", false);
		if (pref) {
			var wWatcher = Cc["@mozilla.org/embedcomp/window-watcher;1"].getService(Ci.nsIWindowWatcher);
			var params = Cc["@mozilla.org/supports-array;1"].createInstance(Ci.nsISupportsArray);
			wWatcher.openWindow(
				null,
				"chrome://obm-extension/content/calendar/sync-dialog.xul",
				"OBM_SYNCHRO",
				"modal,centerscreen,chrome,resizable=no,dialog",
				params
			);
		}
	}
};

/*
 * module registration
 */

var quitObserverModule = {

	mCID: Components.ID("{08495839-ee1f-4758-834e-b9183489af89}"),
	mContractID: "@aliasource.fr/quit-observer;1",

	registerSelf: function (compMgr, fileSpec, location, type) {
		if (!this.registered) {
			compMgr = compMgr.QueryInterface(Ci.nsIComponentRegistrar);
			compMgr.registerFactoryLocation(
				this.mCID,
				"OBM Connector quit observer",
				this.mContractID,
				fileSpec,
				location,
				type
			);

			var catMgr = Cc["@mozilla.org/categorymanager;1"].getService(Ci.nsICategoryManager);
			catMgr.addCategoryEntry("app-startup", this.mContractID, this.mContractID, true, true);

			this.registered = true;
		}
	},

	getClassObject: function (compMgr, cid, iid) {
		if (!cid.equals(this.mCID))
		throw Cr.NS_ERROR_NO_INTERFACE;
		if (!iid.equals(Ci.nsIFactory))
		throw Cr.NS_ERROR_NOT_IMPLEMENTED;
		return this.mFactory;
	},

	mFactory: {
		createInstance: function (outer, iid) {
			if (outer != null)
			throw Cr.NS_ERROR_NO_AGGREGATION;
			return (new QuitObserver()).QueryInterface(iid);
		}
	},

	canUnload: function(compMgr) {
		return true;
	}
};

function NSGetModule(compMgr, fileSpec) {
	return quitObserverModule;
}

